/****************************************************************************

Package:            de.mvcsys.awtutil
Class:              ConsolePane

Task:               Scrollable Console Window with PrintWriter

Author:             Marcus von Cube

Updates:

Datum     Name  Vers.  Bemerkung
-----------------------------------------------------------------------------
26.04.03  MvC   0.1    created
15.06.03  MvC   0.?    ?

****************************************************************************/

package de.mvcsys.awtutil;

import java.awt.*;
import java.io.Writer;
import java.io.PrintWriter;

/**
 *  Scrollable Console Window with Writer
 */
public class ConsolePane extends ScrollPane
{
    /**
     *  Storage for output lines
     */
    private String[] contents;

    /**
     *  Cached width for each line
     */
    private int[] lineWidth;

    /**
     *  Width of the pane
     */
    private int maxWidth;

    /**
     *  Top line (index)
     */
    private int top;

    /**
     *  Line count
     */
    private int length;

    /**
     *  Information about current font
     */
    private FontMetrics metrics;

    /**
     *  Hight of a single line in pixels
     */
    private int lineHeight;

    /**
     *  Y offset of first line
     */
    private int yOffset;


    /**
     *  Class to hold the actual contents
     */
    private class Pane extends Component
    {
        private int width;
        private int height;

        public void paint( Graphics g )
        {
            if ( length == 0 || lineHeight == 0 ) {
                return;
            }
            /*
             *  redraw all visible lines
             */
            Rectangle cb = g.getClipBounds();
            int start = cb.y / lineHeight;
            int end = ( cb.y + cb.height + lineHeight - 1 ) / lineHeight;
            start = start < 0 ? 0 : start;
            end = end > length ? length : end;

            // g.setFont( getFont() );

            for ( int i = start; i < end; ++i ) {
                int y = i * lineHeight;

                g.setColor( getBackground() );
                g.fillRect( 0, y, width, lineHeight );

                String s = contents[ ( i + top ) % contents.length ];
                if ( s != null && s.length() != 0 ) {
                    g.setColor( getForeground() );
                    g.drawString( s, 0, y + yOffset );
                }
            }
        }


        public void update( Graphics g )
        {
            // System.out.println( "u" );
            paint( g );
        }


        public void setPreferredSize( int w, int h )
        {
            if ( width != w || height != h ) {
                width = w;
                height = h;
                invalidate();
            }
        }


        public Dimension getPreferredSize()
        {
            return new Dimension( width, height );
        }


        public Dimension getMinimumSize()
        {
            return new Dimension( width, height );
        }

    };


    /**
     *  Pane instance to hold the actual contents
     */
    private Pane pane;


    /**
     *  Writer for console
     */
    private Writer writer = new Writer() {

        private StringBuffer buffer;
        private boolean flushed = true;
        private char lastChar = 0;

        public void write( char[] cbuf, int off, int len )
        {
            int pos = -1;

            for ( int i = 0; i < len; ++i ) {
                char c = cbuf[ i + off ];
                if ( !Character.isISOControl( c ) ) {
                    /*
                     *  not a control: add to buffer
                     */
                    if ( buffer == null ) {
                        buffer = new StringBuffer( 100 );
                    }
                    flushed = false;
                    buffer.append( c );
                }
                else {
                    /*
                     *  check for line end
                     */
                    if ( c == '\r' ) {
                        flushed = false;
                        flush();
                    }
                    else if ( c == '\n' && lastChar != '\r' ) {
                        flushed = false;
                        flush();
                    }
                }
                lastChar = c;
            }
        }


        public void flush()
        {
            if ( !flushed ) {
                addLine( buffer == null ? null : buffer.toString() );
                flushed = true;
                if ( buffer != null ) {
                    buffer.setLength( 0 );
                }
            }
        }


        public void close()
        {
            flush();
        }
    };


    /**
     *  PrintWriter for convenience
     */
    PrintWriter printWriter;


    /**
     *  Constructor.
     *  @param lines Maximum number of lines stored
     */
    public ConsolePane( int lines )
    {
        super( SCROLLBARS_AS_NEEDED );
        //super( SCROLLBARS_ALWAYS );
        contents = new String[ lines ];
        lineWidth = new int[ lines ];
        pane = new Pane();
        add( pane );
    }


    /**
     *  Don't blank on repaint
     */
    public void update( Graphics g )
    {
        // System.out.println( "U" );
        paint( g );
    }


    /**
     *  set the font and compute new line height and panel size
     */
    public void setFont( Font font )
    {
        super.setFont( font );

        Graphics g = getGraphics();
        if ( g != null ) {
            /*
             *  get font metrics
             */
            metrics = g.getFontMetrics();
            lineHeight = metrics.getHeight();
            yOffset = metrics.getMaxAscent();

            /*
             *  compute width
             */
            maxWidth = 0;
            for ( int i = 0; i < length; ++i ) {
                String s = contents[ ( i + top ) % contents.length ];
                if ( s == null || s.length() == 0 ) {
                    lineWidth[ i ] = 0;
                }
                else {
                    int w = lineWidth[ i ] = metrics.stringWidth( s );
                    if ( w > maxWidth ) {
                        maxWidth = w;
                    }
                }
            }

            /*
             *  scrollbar increment
             */
            Adjustable scrollbar = getVAdjustable();
            if ( scrollbar != null ) {
                scrollbar.setUnitIncrement( lineHeight );
            }
            scrollbar = getHAdjustable();
            if ( scrollbar != null ) {
                scrollbar.setUnitIncrement( 40 );
            }
            pane.setPreferredSize( maxWidth, length * lineHeight );
        }
    }


    /**
     *  Add a line of output
     */
    public synchronized void addLine( String line )
    {
        boolean repaintAll = false;

        if ( metrics == null ) {
            /*
             *  Some intialization required
             */
            setFont( getFont() );
            repaintAll = true;
        }
        int index;
        if ( length == contents.length ) {
            /*
             *  scrolling
             */
            index = top++;
            if ( top == contents.length ) {
                top = 0;
            }
            if ( lineWidth[ index ] == maxWidth ) {
                /*
                 *  recompute width
                 */
                maxWidth = 0;
                lineWidth[ index ] = 0;
                for ( int i = 0; i < length; ++i ) {
                    if ( lineWidth[ i ] > maxWidth ) {
                        maxWidth = lineWidth[ i ];
                    }
                }
            }
            repaintAll = true;
        }
        else {
            /*
             *  put at end
             */
            index = length++;
        }

        /*
         *  store and compute width
         */
        contents[ index ] = line;
        if ( line == null || line.length() == 0 ) {
            lineWidth[ index ] = 0;
        }
        else {
            lineWidth[ index ] = metrics.stringWidth( line );
            if ( lineWidth[ index ] > maxWidth ) {
                maxWidth = lineWidth[ index ];
            }
        }
        int paneHeight = length * lineHeight;
        pane.setPreferredSize( maxWidth, paneHeight );

        /*
         *  refresh output
         */
        validate();

        int vpHeight = getViewportSize().height;
        if ( paneHeight > vpHeight ) {
            /*
             *  make last line visible
             */
            Point pos = getScrollPosition();
            Point newPos = new Point( 0, paneHeight - vpHeight );
            if ( !pos.equals( newPos ) ) {
                setScrollPosition( newPos );
            }
        }

        if ( repaintAll ) {
            pane.repaint( 500 );
        }
        else {
            pane.repaint( 0, paneHeight - lineHeight,
                          maxWidth, lineHeight );
        }
    }


    /**
     *  Get the writer
     */
    public Writer getWriter()
    {
        return writer;
    }


    /**
     *  Get the printWriter
     */
    public PrintWriter getPrintWriter()
    {
        if ( printWriter == null ) {
            printWriter = new PrintWriter( writer );
        }
        return printWriter;
    }


    /**
     *  get the contents array e.g. for dumping to a file
     */
    public String[] getContents()
    {
        String[] result = new String[ length ];
        System.arraycopy( contents, top, result, 0, length - top );
        if ( top != 0 ) {
            /*
             *  copy lines above the top index
             */
            System.arraycopy( contents, 0, result, length - top, top );
        }
        return result;
    }
}



