/****************************************************************************

Package:            de.mvcsys.epocutil
Class:              EpocUtil

Task:               Utilities for EPOC applications
                    EPOC C++ native methods

Author:             Marcus von Cube

Credits:            Inspired by "Professional Symbian Programming"
                    by Martin Tasker, Birmingham 2000

Updates:

Datum     Name  Vers.  Bemerkung
-----------------------------------------------------------------------------
02.05.03  MvC   0.1    created
19.04.05  MvC   0.6    network calls

****************************************************************************/

#include <e32base.h>
#include <coedef.h>
#include <e32keys.h>
#include <eiksvdef.h>
#include <w32std.h>
#include <eikdll.h>
#include <apgtask.h>
#include <nifman.h>
#include <netdial.h>
#include <es_sock.h>
#include <in_sock.h>
#include "de_mvcsys_epocutil_EpocUtil.h"

#ifndef TRUE
#   define TRUE 1
#   define FALSE 0
#endif


/*
 *  container class for private data
 */
class CDllData : public CBase
{
public:
    TInt iReason;
    RWsSession iWsSession;
    RNif iNif;
    RSocketServ iSsSession;
    RSocket iSocket;
    TNifProgressBuf iProgress;

private:
    TInt iWsIsOpen;
    TInt iNifIsOpen;
    TInt iSsIsOpen;

public:
    CDllData();
    ~CDllData();
};

CDllData::CDllData()
{
    /*
     *  connection to window server
     */
    iReason = iWsSession.Connect();
    if ( iReason == KErrNone ) {
        iWsIsOpen = TRUE;
    }

    /*
     *  connection to network interface (NIF) server
     */
    iReason = iNif.Open();
    if ( iReason == KErrNone ) {
        iNifIsOpen = TRUE;
    }

    /*
     *  connection to socket server
     */
    iReason = iSsSession.Connect( 2 );  // # of slots
    if ( iReason == KErrNone ) {
        iSsIsOpen = TRUE;
    }
}

CDllData::~CDllData()
{
    if ( iWsIsOpen ) {
        iWsSession.Close();
    }
    if ( iNifIsOpen ) {
        iNif.Close();
    }
    if ( iSsIsOpen ) {
        iSsSession.Close();
    }
}


/*
 *  get or create the per thread CDllData object
 */
static CDllData *GetData( TInt& aReason )
{
    CDllData *dllData = REINTERPRET_CAST( CDllData *, Dll::Tls() );
    if ( dllData == NULL ) {
        /*
         *  new object
         */
        TRAP( aReason, dllData = new( ELeave ) CDllData; );
        /*
         *  store pointer for later reuse
         */
        Dll::SetTls( dllData );
    }
    if ( dllData != NULL ) {
        aReason = dllData->iReason;
    }
    return aReason == KErrNone ? dllData : NULL;
}


/*
 * Class:     de_mvcsys_epocutil_EpocUtil
 * Method:    eCleanup
 * Signature: ()I
 *
 * clean up tls data when thread ends
 */
JNIEXPORT void JNICALL Java_de_mvcsys_epocutil_EpocUtil_eCleanup
  ( JNIEnv * /* jEnv */, jclass /* jClass */ )
{
    CDllData *dllData = REINTERPRET_CAST( CDllData *, Dll::Tls() );
    if ( dllData != NULL ) {
        delete dllData;
        Dll::FreeTls();
    }
}


/*
 *  helper to get a java string into a HBufC
 */
static HBufC *MakeStringLC( JNIEnv *jEnv, jstring jString )
{
    TInt len = jString == NULL ? 0 : jEnv->GetStringLength( jString );
    HBufC *result = HBufC::NewMaxLC( len );

    if ( len > 0 ) {
        const jchar *src = jEnv->GetStringChars( jString, JNI_FALSE );
        TPtr dest = result->Des();

        for ( int i = 0; i < len; ++i ) {
            dest[ i ] = STATIC_CAST( const TText, src[ i ] );
        }
        jEnv->ReleaseStringChars( jString, src );
    }
    return result;
}


#ifdef _UNICODE
/*
 *  helper to get a java string into a HBufC8
 */
static HBufC8 *MakeStringLC8( JNIEnv *jEnv, jstring jString )
{
    TInt len = jString == NULL ? 0 : jEnv->GetStringLength( jString );
    HBufC8 *result = HBufC8::NewMaxLC( len );

    if ( len > 0 ) {
        const jchar *src = jEnv->GetStringChars( jString, JNI_FALSE );
        TPtr8 dest = result->Des();

        for ( int i = 0; i < len; ++i ) {
            dest[ i ] = STATIC_CAST( const TText8, src[ i ] );
        }
        jEnv->ReleaseStringChars( jString, src );
    }
    return result;
}
#else
/*
 *  On non-unicode systems, this is the same as MakeStringLC
 */
#define MakeStringLC8 MakeStringLC
#endif


/*
 *  helper to create a Java string from a TdesC
 */
static jstring MakeJStringL( JNIEnv *jEnv, TDesC *aDes )
{
    if ( aDes == NULL ) {
        return NULL;
    }
#ifdef _UNICODE
    return jEnv->NewString( STATIC_CAST( const jchar *, aDes->Ptr() ),
                            STATIC_CAST( jsize, aDes->Length() ) );
#else
    TInt len = aDes->Length();
    if ( len == 0 ) {
        static const jchar dummy = 0;
        return jEnv->NewString( &dummy, 0 );
    }
    HBufC16 *buffer = HBufC16::NewMaxLC( len );
    TPtr16 buf = buffer->Des();
    for ( int i = 0; i < len; ++i ) {
        buf[ i ] = STATIC_CAST( const TText16, (*aDes)[ i ] );
    }
    jstring result = jEnv->NewString( STATIC_CAST( const jchar *, buf.Ptr() ),
                                      STATIC_CAST( jsize, len ) );
    CleanupStack::PopAndDestroy( 1 );
    return result;
#endif
}


/*
 *  Make a jint from a thread id
 */
inline jint Tid2jint( TThreadId& aTid )
{
    return *( (jint *) &aTid );
}


/*
 *  Make a jint from a thread id
 */
inline TThreadId Jint2tid( jint jTid )
{
    return *( (TThreadId *) &jTid );
}


/*
 *  Store the TRAP reason in class
 */
static void SetReason( JNIEnv *jEnv, jclass jClass, TInt aReason )
{
    jEnv->SetStaticIntField( jClass,
                             jEnv->GetStaticFieldID( jClass, "reason", "I" ),
                             STATIC_CAST( jint, aReason ) );
}


/*
 *  Display the task list
 */
static TInt DisplayTaskList()
{
    TInt reason;
    CDllData *dllData = GetData( reason );
    if ( dllData == NULL ) {
        return reason;
    }
    TEikServEvent taskListMsg = EEikServShowTaskList;
    TWsEvent event;
    event.SetType( EEventUser );
    Mem::Copy( event.EventData(), &taskListMsg, sizeof( taskListMsg ) );
    TInt wgId = (dllData->iWsSession).FindWindowGroupIdentifier(
                                            0, __EIKON_SERVER_NAME, 0 );
    (dllData->iWsSession).SendEventToWindowGroup( wgId, event );
    return KErrNone;
}


/*
 * Class:     de_mvcsys_epocutil_EpocUtil
 * Method:    eOpenTaskList
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_de_mvcsys_epocutil_EpocUtil_eDisplayTaskList
  ( JNIEnv * jEnv, jclass jClass )
{
    TInt reason = DisplayTaskList();
    if ( reason != KErrNone ) {
        SetReason( jEnv, jClass, reason );
        return -1;
    }
    return 0;
}


/*
 *  Start an Application
 */
static jint StartAppL( JNIEnv *jEnv, jstring jLibName,
                       jstring jDocName, jstring jTailEnd, jint jCmd )
{
    CApaCommandLine *cmdLine = CApaCommandLine::NewLC();
    HBufC *libName = MakeStringLC( jEnv, jLibName );
    HBufC *docName = MakeStringLC( jEnv, jDocName );
    HBufC8 *tailEnd = MakeStringLC8( jEnv, jTailEnd );
    TApaCommand cmd = STATIC_CAST( TApaCommand, jCmd );

    if ( libName->Length() > 0 ) {
        cmdLine->SetLibraryNameL( *libName );
    }
    if ( docName->Length() > 0 ) {
        cmdLine->SetDocumentNameL( *docName );
    }
    if ( tailEnd->Length() > 0 ) {
        cmdLine->SetTailEndL( *tailEnd );
    }
    cmdLine->SetCommandL( cmd );
    CleanupStack::PopAndDestroy( 3 ); // libName, docName, tailEnd

    TThreadId tid = EikDll::StartAppL( *cmdLine );
    CleanupStack::PopAndDestroy();  // cmdLine

    return Tid2jint( tid );
}


/*
 * Class:     de_mvcsys_epocutil_EpocUtil
 * Method:    eStartApp
 * Signature: (Ljava/lang/String;)I
 */
JNIEXPORT jint JNICALL Java_de_mvcsys_epocutil_EpocUtil_eStartApp
  ( JNIEnv *jEnv, jclass jClass, jstring jLibName, jstring jDocName,
    jstring jTailEnd, jint jCmd )
{
    TInt result = 0;
    TRAPD( reason, result = StartAppL( jEnv,
                                       jLibName, jDocName, jTailEnd, jCmd ) );
    if ( reason != KErrNone ) {
        SetReason( jEnv, jClass, reason );
        return 0;
    }
    return STATIC_CAST( jint, result );
}


/*
 *  open a document
 */
static void OpenDocumentL( JNIEnv *jEnv, jstring jDocName )
{
    HBufC *docName = MakeStringLC( jEnv, jDocName );
    EikDll::StartDocL( *docName );
    CleanupStack::PopAndDestroy( 1 );
}


/*
 * Class:     de_mvcsys_epocutil_EpocUtil
 * Method:    eOpenDocument
 * Signature: (Ljava/lang/String;)I
 */
JNIEXPORT void JNICALL Java_de_mvcsys_epocutil_EpocUtil_eOpenDocument
  ( JNIEnv *jEnv, jclass jClass, jstring jDocName )
{
    TRAPD( reason, OpenDocumentL( jEnv, jDocName ) );
    if ( reason != KErrNone ) {
        SetReason( jEnv, jClass, reason );
    }
}


/*
 *  Wait for a thread
 */
static TInt WaitThread( const TThreadId aTid )
{
    RThread thread;
    TRequestStatus status;
    
    thread.Open( aTid );
    thread.Logon( status );
    User::WaitForRequest( status );
    return status.Int();
}


/*
 * Class:     de_mvcsys_epocutil_EpocUtil
 * Method:    eWaitThread
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_de_mvcsys_epocutil_EpocUtil_eWaitThread
  ( JNIEnv * /* jEnv */, jclass /* jClass */, jint jTid )
{
    return STATIC_CAST( jint, WaitThread( Jint2tid( jTid ) ) );
}


/*
 *  Get window group from thread id
 */
static TInt WindowGroupByThreadId( TInt aLastId, TThreadId& aTid )
{
    TInt reason;
    CDllData *dllData = GetData( reason );
    if ( dllData == NULL ) {
        return reason;
    }
    return (dllData->iWsSession).FindWindowGroupIdentifier( aLastId, aTid );
}


/*
 * Class:     de_mvcsys_epocutil_EpocUtil
 * Method:    eWindowGroupByThread
 * Signature: (II)I
 */
JNIEXPORT jint JNICALL Java_de_mvcsys_epocutil_EpocUtil_eWindowGroupByThread
  ( JNIEnv * /* jEnv */, jclass /* jClass */, jint jLastId, jint jTid )
{
    TThreadId tid;
    if ( jTid == 0 ) {
        tid = Jint2tid( jTid );
    }
    else {
        tid = RThread().Id();
    }
    return STATIC_CAST( jint,
            WindowGroupByThreadId( STATIC_CAST( TInt, jLastId ), tid ) );
}
    

/*
 *  find named window group
 */
static TInt WindowGroupByNameL( JNIEnv *jEnv,
                                TInt aLastId, jstring jName, TInt aOffset )
{
    TInt reason;
    CDllData *dllData = GetData( reason );
    if ( dllData == NULL ) {
        return reason;
    }
    HBufC *name = MakeStringLC( jEnv, jName );
    TInt wgId = (dllData->iWsSession).FindWindowGroupIdentifier( aLastId,
                                                                 *name,
                                                                 aOffset );
    CleanupStack::PopAndDestroy( 1 );
    return wgId; 
}


/*
 * Class:     de_mvcsys_epocutil_EpocUtil
 * Method:    eWindowGroupByName
 * Signature: (ILjava/lang/String;)I
 */
JNIEXPORT jint JNICALL Java_de_mvcsys_epocutil_EpocUtil_eWindowGroupByName
  ( JNIEnv *jEnv, jclass /* jClass */,
    jint jLastId, jstring jTaskName, jint jOffset )
{
    TInt wgId = KErrNotFound;
    TRAPD( reason, wgId = WindowGroupByNameL( jEnv,
                                              STATIC_CAST( TInt, jLastId ),
                                              jTaskName,
                                              STATIC_CAST( TInt, jOffset )
                                            )
         );
    if ( reason != KErrNone ) {
        return STATIC_CAST( jint, reason );
    }
    return STATIC_CAST( jint, wgId );
}
    

/*
 *  get the window group id which has focus
 */
static TInt WindowGroupWithFocus()
{
    TInt reason;
    CDllData *dllData = GetData( reason );
    if ( dllData == NULL ) {
        return reason;
    }
    return (dllData->iWsSession).GetFocusWindowGroup();
}


/*
 * Class:     de_mvcsys_epocutil_EpocUtil
 * Method:    eWindowGroupWithFocus
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_de_mvcsys_epocutil_EpocUtil_eWindowGroupWithFocus
  ( JNIEnv * /* jEnv */, jclass /* jClass */ )
{
    return STATIC_CAST( jint, WindowGroupWithFocus() );
}


/*
 *  get the name of a window group
 */
static jstring GetWindowGroupNameL( JNIEnv *jEnv, TInt aWgId )
{
    TInt reason;
    CDllData *dllData = GetData( reason );
    if ( dllData == NULL ) {
        User::Leave( reason );
    }
    HBufC *buffer = HBufC::NewMaxLC( 256 );
    TPtr name = buffer->Des();
    TInt err = (dllData->iWsSession).GetWindowGroupNameFromIdentifier( aWgId,
                                                                       name );
    User::LeaveIfError( err );
    jstring result = MakeJStringL( jEnv, &name );
    CleanupStack::PopAndDestroy( 1 );
    return result;
}


/*
 * Class:     de_mvcsys_epocutil_EpocUtil
 * Method:    eGetWindowGroupName
 * Signature: (I)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_de_mvcsys_epocutil_EpocUtil_eGetWindowGroupName
  ( JNIEnv *jEnv, jclass jClass, jint jWgId )
{
    jstring result = NULL;
    TRAPD( reason,
           result = GetWindowGroupNameL( jEnv,
                                         STATIC_CAST( TInt, jWgId ) ) );
    SetReason( jEnv, jClass, reason );
    return result;
}


/*
 *  set the window group ordinal position
 */
static TInt SetWindowGroupPosition( TInt aWgId, TInt aPos )
{
    TInt reason;
    CDllData *dllData = GetData( reason );
    if ( dllData == NULL ) {
        return reason;
    }
    return (dllData->iWsSession).SetWindowGroupOrdinalPosition( aWgId, aPos );
}


/*
 * Class:     de_mvcsys_epocutil_EpocUtil
 * Method:    eSetWindowGroupPosition
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_de_mvcsys_epocutil_EpocUtil_eSetWindowGroupPosition
  (JNIEnv * /* jEnv */, jclass /* jClass */, jint jWgId, jint jPos )
{
    return STATIC_CAST( jint, 
                        SetWindowGroupPosition( STATIC_CAST( TInt, jWgId ),
                                                STATIC_CAST( TInt, jPos ) )
                      );
}


/*
 * Helper: Dial a connection and get local IP address
 */
static TInt NetDialL()
{
    TInt reason;

    CDllData *dllData = GetData( reason );
    if ( dllData == NULL ) {
        User::Leave( reason );
    }

    /*
     *  dial
     */
    RNif& nif = dllData->iNif;
    RSocket& socket = dllData->iSocket;
    TNifProgressBuf& progress = dllData->iProgress;
    TRequestStatus request, request2;

    TBool connected = FALSE;
    nif.NetworkActive( connected );
    if ( !connected ) {
        nif.Progress( progress() );
        connected = progress().iStage >= EConnectionOpen;
    }

    socket.Open( dllData->iSsSession, KAfInet, KSockDatagram, 
                 KProtocolInetUdp );

    TBool first = TRUE;
    _LIT8( dummy, "EpocUtil" );
    TSockXfrLength sent;

    while ( !connected ) {
        nif.ProgressNotification( progress, request );
        if ( first ) {
            /*
             *  send a dummy packet to force a dial out
             */
            TInetAddr myAddr( 33333 );
            TInetAddr destAddr( 0xc0a80101, 33333 );
            reason = socket.Bind( myAddr );
            if ( reason != KErrNone ) {
                nif.CancelProgressNotification();
                socket.Close();
                User::Leave( reason );
            }
            socket.SendTo( dummy, destAddr, 0, request2, sent );
            first = FALSE;
        }
        User::WaitForRequest( request, request2 );
        if ( request2.Int() != KRequestPending ) {
            reason = request2.Int();
            if ( reason == KErrNone ) {
                /*
                 *  Send complete
                 */
                nif.CancelProgressNotification();
                connected = TRUE;
                continue;
            }
        }
        else {
            /*
             *  Progress Info
             */
            reason = request.Int();
            if ( reason == KErrNone ) {
                reason = progress().iError;
            }
            if ( reason != KErrNone ) {
                socket.CancelSend();
            }
        }
        if ( reason != KErrNone ) {
            socket.Close();
            User::Leave( reason );
        }
        nif.NetworkActive( connected );
        if ( !connected ) {
            connected = progress().iStage >= EConnectionOpen;
        }
    }
    socket.CancelSend();

    /*
     *  get Interface information
     */
    reason = socket.SetOpt( KSoInetEnumInterfaces, KSolInetIfCtrl );
    if ( reason != KErrNone ) {
        socket.Close();
        User::Leave( reason );
    }

#   define LOCALHOST 0x7f000001
    TInt result = 0;
    do {
        TPckgBuf<TSoInetInterfaceInfo> info;
        reason = socket.GetOpt( KSoInetNextInterface, KSolInetIfCtrl, info );
        if ( reason != KErrNone ) {
            socket.Close();
            User::Leave( reason );
        }
        result = info().iAddress.Address();
        if ( result == 0 ) {
            /*
             *  resort to gateway if address is unknown
             */
            result = info().iDefGate.Address();
        }
    } while ( result == LOCALHOST );

    socket.Close();

    return result;
}


/*
 * Class:     de_mvcsys_epocutil_EpocUtil
 * Method:    eNetDial
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_de_mvcsys_epocutil_EpocUtil_eNetDial
  (JNIEnv * jEnv, jclass jClass)
{
    jint result = 0;
    TRAPD( reason, result = STATIC_CAST( jint, NetDialL() ) );
    if ( reason != KErrNone ) {
        SetReason( jEnv, jClass, reason );
    }
    return result;
}


/*
 * Helper: Hang up a connection
 */
static void NetStopL()
{
    TInt reason;

    CDllData *dllData = GetData( reason );
    if ( dllData == NULL ) {
        User::Leave( reason );
    }
    RNif& nif = dllData->iNif;
    User::LeaveIfError( nif.Stop() );
}


/*
 * Class:     de_mvcsys_epocutil_EpocUtil
 * Method:    eNetStop
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_de_mvcsys_epocutil_EpocUtil_eNetStop
  (JNIEnv *jEnv, jclass jClass)
{
    TRAPD( reason, NetStopL() );
    if ( reason != KErrNone ) {
        SetReason( jEnv, jClass, reason );
    }
}
