/****************************************************************************

Package:            de.mvcsys.epocutil
Class:              EpocUtil

Task:               Utilities for EPOC applications

Author:             Marcus von Cube

Credits:            Inspired by "Professional Symbian Programming"
                    by Martin Tasker, Birmingham 2000

Updates:

Datum     Name  Vers.  Bemerkung
-----------------------------------------------------------------------------
02.05.03  MvC   0.1    created
25.05.03  MvC   0.4    window groups
19.04.05  MvC   0.6    network calls

****************************************************************************/

package de.mvcsys.epocutil;

import java.util.*;
import java.io.*;
import java.net.*;


/**
 *  <p>Utility function for EPOC applications.</p>
 *
 *  <p>On EPOC needs EpocUtil.dll in \System\libs on any drive.
 *  On other machines, most functions are emulated.</p>
 *
 *  <p>Hint: you can create a file <code>epocutil.properties</code> somewhere
 *  in the classpath (default is \System\Java\classes on any drive).
 *  Here you can change or add system properties not properly maintained by
 *  EPOC Java, such as language, region or timezone. Central European users
 *  should set <code>user.timezone=ECT</code>, otherwise GMT is used.</p>
 */
public class EpocUtil
{
    /**
     *  Version
     */
    public static final String VERSION = "Version 0.60, 20 Apr 2005";

    /**
     *  command line command to open an existing document (0)
     */
    public static final int CMD_OPEN = 0;

    /**
     *  command line command to create a new document (1)
     */
    public static final int CMD_CREATE = 0;

    /**
     *  command line command to run the application (2)
     */
    public static final int CMD_RUN = 0;

    /**
     *  command line command to run the application in the background (3)
     */
    public static final int CMD_BACKGROUND = 0;

    /**
     *  default offset for window group names (12)
     */
    public static final int WG_NAME_OFFSET = 12;

    /**
     *  last reason code
     */
    public static int reason;

    /**
     *  window identifier, should be set after AWT is initialized
     */
    private static int window;

    /*
     *  private member variables
     */
    private static boolean isEpoc;
    private static boolean epocLibInstalled;
    private static boolean debug = false;
    private static final Object lock = new Object();

    /**
     *  Make static by hiding the constructor
     */
    private EpocUtil()
    {
    }


    /*
     *  Static initialization
     */
    static {
        /*
         *  Load additional system properties
         */
        try {
            Properties props = new Properties();
            InputStream in =
                ClassLoader.getSystemResourceAsStream( "EpocUtil.properties" );
            if ( in == null ) {
                if ( debug ) {
                    System.out.println( "no props, classpath="
                            + System.getProperty( "java.class.path" ) );
                }
            }
            else {
                props.load( in );
                in.close();
            }

            String debugProp =
                  props.getProperty( "debug", System.getProperty( "debug" ) );
            if ( debugProp != null ) {
                debug = debugProp.equals( "on" )
                     || debugProp.equals( "true" )
                     || debugProp.equals( "1" );
            }

            /*
             *  copy the properties to System
             */
            Properties sysProps = System.getProperties();
            Enumeration names = props.propertyNames();
            while ( names.hasMoreElements() ) {
                String name = (String) names.nextElement();
                String value = props.getProperty( name );
                if ( debug ) {
                    System.out.print( name + "=" + value );
                    String oldValue = sysProps.getProperty( name );
                    if ( oldValue != null ) {
                        System.out.print( " (" + oldValue + ")" );
                    }
                    System.out.println();
                }
                sysProps.put( name, value );
            }
            if ( debug ) {
                sysProps.list( System.out );
            }
        }
        catch ( Exception e ) {
            if ( debug ) {
                e.printStackTrace();
            }
        }
        try {
            TimeZone tz =
                TimeZone.getTimeZone( System.getProperty( "user.timezone" ) );
            TimeZone.setDefault( tz );
        }
        catch ( Exception e ) {
            if ( debug ) {
                e.printStackTrace();
            }
        }

        try {
            Locale loc = new Locale( System.getProperty( "user.language" ),
                                     System.getProperty( "user.region" ) );
            Locale.setDefault( loc );
        }
        catch ( Exception e ) {
            if ( debug ) {
                e.printStackTrace();
            }
        }

        if ( debug ) {
            System.out.println( "TimeZone: " + TimeZone.getDefault().getID() );
            System.out.println( "Locale: " + Locale.getDefault() );
        }

        /*
         *  Load DLL if running on EPOC
         */
        try {
            isEpoc = System.getProperty( "os.name" ).startsWith( "EPOC" );
            if ( isEpoc ) {
                System.loadLibrary( "EpocUtil" );
                epocLibInstalled = true;
            }
        }
        catch ( Exception e ) {
            if ( debug ) {
                e.printStackTrace();
            }
        }

    }


    /**
     *  check if running on an EPOC machine
     */
    public static boolean isEpoc()
    {
        return isEpoc;
    }


    /**
     *  check if debugging desired
     */
    public static boolean isDebug()
    {
        return debug;
    }


    /**
     *  thread cleanup, call at end of run()
     */
    public static void cleanup()
    {
        if ( isEpoc && epocLibInstalled ) {
            eCleanup();
        }
    }


    /**
     *  set the id of our window (current focus or app name)
     */
    public static void setWindow()
    {
        window = windowGroupWithFocus();
        String name = getWindowGroupName();
        boolean delay = false;

        if ( stripWindowGroupName( name ).equals( "STDOUT" ) ) {
            /*
             *  we do not want the console!
             */
            int w = windowGroup();
            if ( w >= 0 ) {
                window = w;
                delay = true;
            }
            else if ( debug ) {
                /*
                 *  show all window groups
                 */
                for ( int wgId = 0; wgId != -1; ) {
                    wgId = windowGroupByName( wgId, "*", 0 );
                    if ( wgId != -1 ) {
                        getWindowGroupName( wgId );
                    }
                }
            }
        }
        if ( debug ) {
            System.out.println( "Window(" + window + "): \""
                                + stripWindowGroupName() + "\"" );
            if ( delay ) {
                try {
                    Thread.sleep( 1500 );
                }
                catch ( Exception e ) {}
            }
        }
        /*
         *  we want our window to pop up!
         */
        foreground();
    }


    /**
     *  get the stored window handle
     */
    public static int getWindow()
    {
        return window;
    }


    /**
     *  open the system task list
     */
    public static void displayTaskList()
    {
        if ( isEpoc && epocLibInstalled ) {
            eDisplayTaskList();
        }
    }


    /**
     *  Start an application.
     *
     *  @param libName application name, e. g. "Word"
     *  @param background if <code>true</code> app is run in the
     *                    background
     *  @return EPOC thread identifier
     */
    public static int startApp( String libName, boolean background )
            throws EpocException, IOException
    {
        return startApp( libName, null, null,
                         background ? CMD_BACKGROUND : CMD_RUN );
    }


    /**
     *  Start an application with document.
     *
     *  @param libName application name, e. g. "Word"
     *  @param docName document name, e. g. "c:\\Documents\Important Text"
     *  @return EPOC thread identifier
     */
    public static int startApp( String libName, String docName )
            throws EpocException, IOException
    {
        return startApp( libName, docName, null, CMD_OPEN );
    }


    /**
     *  Start an application with complete parameter list.
     *
     *  @param libName application name, e. g. "Word"
     *  @param docName document name, e. g. "c:\\Documents\Important Text"
     *  @param tailEnd more parameters, e. g. the OPO for OPL programs
     *  @param cmd one of CMD_OPEN, CMD_CREATE, CMD_RUN or CMD_BACKGROUND
     *  @return EPOC thread identifier
     */
    public static int startApp( String libName, String docName,
                                String tailEnd, int cmd )
            throws EpocException, IOException
    {
        if ( isEpoc && epocLibInstalled ) {
            synchronized ( lock ) {
                reason = 0;
                int result = eStartApp( libName, docName, tailEnd, cmd );
                checkReason();
                return result;
            }
        }
        else {
            Runtime.getRuntime().exec( "\"" + libName + "\" \""
                                       + docName + "\"" + tailEnd );
            return 0;
        }
    }


    /**
     *  open a document
     */
    public static void openDocument( String docName )
            throws EpocException, IOException
    {
        String extension = "";
        int pos = docName.lastIndexOf( "." );
        if ( pos > 0 ) {
            extension = docName.substring( pos + 1 ).toLowerCase();
        }

        if ( isEpoc && epocLibInstalled ) {
            if ( extension.equals( "htm" ) || extension.equals( "html" ) ) {
                reason = 0;
                synchronized ( lock ) {
                    reason = 0;
                    debug( "opera" );
                    eStartApp( "opera", docName, null, CMD_OPEN );
                    debug( "opera, reason = " + reason );
                    if ( reason != 0 ) {
                        reason = 0;
                        debug( "Web" );
                        eStartApp( "Web", docName, null, CMD_OPEN );
                        debug( "Web, reason = " + reason );
                    }
                    checkReason();
                }
                return;
            }

            synchronized ( lock ) {
                reason = 0;
                eOpenDocument( new File( docName ).getCanonicalPath() );
                checkReason();
            }
        }
        else {
            /*
             *  this works on some operating systems but not on all
             */
            String opsys = System.getProperty( "os.name" ).toLowerCase();
            if ( opsys.startsWith( "os/2" ) ) {
                if ( extension.equals( "htm" ) || extension.equals( "html" ) )
                {
                    docName = "firefox " + docName;
                }
                else {
                    docName = "e " + docName;
                }
            }
            Runtime.getRuntime().exec( docName );
        }
    }


    /**
     *  Wait for a thread to complete
     *  @param the id returned by one of the start or open methods
     */
    public static int waitThread( int id )
    {
        if ( id != 0 && isEpoc && epocLibInstalled ) {
            return eWaitThread( id );
        }
        return -1;
    }


    /**
     *  Get the window id which has keyboard focus
     */
    public static int windowGroupWithFocus()
    {
        if ( isEpoc && epocLibInstalled ) {
            return eWindowGroupWithFocus();
        }
        return -1;
    }


    /**
     *  Get a thread's window group id
     *  @param tid the id returned by one of the start or open methods
     */
    public static int windowGroupByThread( int lastId, int tid )
    {
        if ( tid != 0 && isEpoc && epocLibInstalled ) {
            return eWindowGroupByThread( lastId, tid );
        }
        return -1;
    }


    /**
     *  Get a window group id from its name
     *  @param lastId the last returned id or 0
     *  @param name the name of the window group
     *  @param offset number of characters to skip at the beginning
     */
    public static int windowGroupByName( int lastId, String name, int offset )
    {
        if ( isEpoc && epocLibInstalled ) {
            int wgId = -1;
            if ( name != null ) {
                wgId = eWindowGroupByName( lastId, name, offset );
            }
            if ( debug && false ) {
                System.out.println( "name: \"" + name
                                    + "\", offset: " + offset );
                name = getWindowGroupName( wgId );
                System.out.println( "found(" + wgId + "): \"" + name + "\"" );
            }
            return wgId;
        }
        return -1;
    }


    /**
     *  Get a window group id from its name.
     *  The offset is set to the default value.
     *  @param lastId the last returned id or 0
     *  @param name the name of the window group
     */
    public static int windowGroupByName( int lastId, String name )
    {
        return windowGroupByName( lastId, name, WG_NAME_OFFSET );
    }


    /**
     *  Get the window group from the system property
     *  <code>com.symbian.appName</code>.
     *  The offset is set to the default offset of 12.
     *  @param lastId the last returned id or 0
     */
    public static int windowGroup( int lastId )
    {
        String name = System.getProperty( "com.symbian.appName" );
        if ( name == null ) {
            return -1;
        }
        return windowGroupByName( lastId, name + '\0', WG_NAME_OFFSET );
    }


    /**
     *  Get the window group from the system property
     *  <code>com.symbian.appname</code>.
     *  The offset is set to the default offset of 12.
     *  @param lastId the last returned id or 0
     */
    public static int windowGroup()
    {
        return windowGroup( 0 );
    }


    /**
     *  Get the name of a window group
     */
    public static String getWindowGroupName( int wgId )
    {
        if ( isEpoc && epocLibInstalled ) {
            String name = eGetWindowGroupName( wgId );
            if ( debug && name != null ) {
                System.out.println( wgId + ": \""
                                    + name.replace( (char) 0, (char) 0x95 )
                                    + "\", len = "
                                    + name.length() );
                if ( false ) {
                    for ( int i = 0, l = name.length(); i < l; ++i ) {
                        System.out.print( " " + ( (int) name.charAt( i ) ) );
                    }
                    System.out.println();
                }
            }
            return name;
        }
        return null;
    }


    /**
     *  Get the name of our window group
     */
    public static String getWindowGroupName()
    {
        return getWindowGroupName( window );
    }


    /**
     *  Strip the window group name to the display form
     */
    public static String stripWindowGroupName( String name )
    {
        if ( name.length() < WG_NAME_OFFSET ) {
            return "";
        }
        int tail;
        for ( tail = name.length() - 1;
              tail >= WG_NAME_OFFSET && name.charAt( tail ) == (char) 0;
              --tail );
        String result = name.substring( WG_NAME_OFFSET, tail + 1 );
        if ( result.indexOf( (char) 0 ) > 0 ) {
            result = result.replace( (char) 0, ' ' );
        }
        return result;
    }


    /**
     *  Get the stripped window group name of a window
     */
    public static String stripWindowGroupName( int wgId )
    {
        return stripWindowGroupName( getWindowGroupName( wgId ) );
    }


    /**
     *  Get the stripped window group name of our window
     */
    public static String stripWindowGroupName()
    {
        return stripWindowGroupName( window );
    }


    /**
     *  Set the window group position
     *  @param id the window group id
     *  @param pos position (0 = foreground, -1 = background)
     */
    public static int setWindowGroupPosition( int wgId, int pos )
    {
        if ( wgId >= 0 && isEpoc && epocLibInstalled ) {
            return eSetWindowGroupPosition( wgId, pos );
        }
        return -1;
    }


    /**
     *  Bring a window group to the foreground
     *  @param id the window group id
     */
    public static int foregroundWindowGroup( int wgId )
    {
        return setWindowGroupPosition( wgId, 0 );
    }


    /**
     *  Bring myself to the foreground
     */
    public static int foreground()
    {
        return foregroundWindowGroup( window );
    }


    /**
     *  Bring a window group to the background
     *  @param id the window group id
     */
    public static int backgroundWindowGroup( int wgId )
    {
        return setWindowGroupPosition( wgId, -1 );
    }


    /**
     *  Bring myself to the foreground
     */
    public static int background()
    {
        return foregroundWindowGroup( window );
    }


    /**
     *  Issue a dial out request
     *  @return local address of interface
     */
    public static InetAddress netDial()
            throws EpocException, UnknownHostException
    {
        if ( isEpoc && epocLibInstalled ) {
            int result;
            synchronized ( lock ) {
                reason = 0;
                result = eNetDial();
                checkReason();
            }
            String dotted = ( ( result >> 24 ) & 255 ) + "."
                          + ( ( result >> 16 ) & 255 ) + "."
                          + ( ( result >>  8 ) & 255 ) + "."
                          + ( result & 255 );
            return InetAddress.getByName( dotted );
        }
        else {
            return InetAddress.getLocalHost();
        }
    }


    /**
     *  Hang up connection
     */
    public static void netStop()
            throws EpocException
    {
        if ( isEpoc && epocLibInstalled ) {
            synchronized ( lock ) {
                reason = 0;
                eNetStop();
                checkReason();
            }
        }
    }


    /**
     *  Throw Exception if reason != 0
     */
    private static void checkReason()
        throws EpocException
    {
        if ( reason != EpocException.KErrNone ) {
            throw new EpocException( reason );
        }
    }


    /**
     *  Debug output to the console
     */
    public static void debug( Object o )
    {
        if ( isDebug() ) {
            System.out.println( o );
        }
    }


    /*
     *  native methods, only available on EPOC
     */
    private static native void   eCleanup();
    private static native int    eDisplayTaskList();
    private static native int    eStartApp( String libName, String docName,
                                            String tailEnd, int cmd );
    private static native void   eOpenDocument( String docName );
    private static native int    eWaitThread( int tid );
    private static native int    eWindowGroupByThread( int lastId, int tid );
    private static native int    eWindowGroupByName( int lastId, String name,
                                                     int offset );
    private static native int    eWindowGroupWithFocus();
    private static native String eGetWindowGroupName( int wgid );
    private static native int    eSetWindowGroupPosition( int wgid, int pos );

    private static native int    eNetDial();
    private static native void   eNetStop();

}
