/****************************************************************************

Package:            de.mvcsys.smtpauth
Class:              SmtpAuthGui

Task:               GUI for SmtpAuth

Author:             Marcus von Cube

Updates:

Datum     Name  Vers.  Bemerkung
-----------------------------------------------------------------------------
26.04.03  MvC   0.1    created
25.05.03  MvC   0.42   new EpocUtil features
04.05.03  MvC   0.5    POP
20.04.05  MvC   0.8    Disconnect

****************************************************************************/

package de.mvcsys.smtpauth;

import de.mvcsys.epocutil.*;
import de.mvcsys.awtutil.*;
import de.mvcsys.util.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

/**
 *  Main Class
 */
public class SmtpAuthGui extends EFrame
{
    public PrintWriter out;
    public ConsolePane console;
    public SmtpAuth engine;
    public Thread thread;
    public String[] args;
    public MenuItem restartItem;
    public MenuItem disconnectItem;
    public MenuItem saveItem;
    public MenuItem settingsItem;
    public MenuItem emailItem;
    public MenuItem sinkItem;
    public MenuItem toolbar5;
    public MenuItem toolbar6;

    /**
     *  Build the GUI and start the engine
     */
    public SmtpAuthGui( String[] args )
    {
        super( SmtpAuth.TITLE );
        String[] buttons;

        if ( EpocUtil.isDebug() ) {
            buttons = new String[] { "Exit", "Restart", "Settings",
                                     getEmailTitle(), "5", "6" };
        }
        else if ( isEpoc ) {
            buttons = new String[] { "Exit", "Restart", "Settings",
                                     getEmailTitle() };
        }
        else {
            buttons = new String[] { "Exit", "Restart", "Settings" };
        }
        addToolbar( SmtpAuth.NAME, buttons );

        aboutDialog.text = new String[] {
            "--" + SmtpAuth.TITLE + "--",
            "Allows your mail client to use SMTP",
            "servers which require authentication",
            "with user id and password.",
            SmtpAuth.VERSION,
            SmtpAuth.AUTHOR
        };
        aboutDialog.setTitle( "About " + SmtpAuth.NAME );
        setHelpDocument( "SmtpAuth.html" );

        console = new ConsolePane( 100 );
        add( console, BorderLayout.CENTER );
        out = console.getPrintWriter();

        setVisible( true );
        show();

        this.args = args;
        startUp();
    }


    /**
     *  Add our menu items
     */
    public void addUserItems( Menu m )
    {
        debug( "addUserItems" );
        if ( m == fileMenu ) {
            restartItem = addMenuItem( m, "Restart", KeyEvent.VK_R );
            saveItem = addMenuItem( m, "Save Screen", KeyEvent.VK_S, true );
            if ( isEpoc ) {
                disconnectItem = addMenuItem( m, "Disconnect from Internet",
                                              KeyEvent.VK_U );
            }
            m.addSeparator();
        }
        else if ( m == toolsMenu ) {
            settingsItem = addMenuItem( m, "Settings", KeyEvent.VK_S );
            String emailTitle = getEmailTitle();
            if ( emailTitle != null ) {
                emailItem = addMenuItem( m, emailTitle, KeyEvent.VK_M );
            }
            if ( isEpoc ) {
                sinkItem = addMenuItem( m, "Activate \"Sink\"",
                                        KeyEvent.VK_I );
            }
            if ( EpocUtil.isDebug() ) {
                toolbar5 = addMenuItem( m, "5", KeyEvent.VK_5 );
                toolbar6 = addMenuItem( m, "6", KeyEvent.VK_6 );
            }
            m.addSeparator();
        }
    }


    /**
     *  Action Listener.
     *  A sub class of EFrame that implements actionPerformed should call
     *  super.actionPerformed( e ), to ensure that the actions handled by
     *  EFrame are not ignored.
     */
    public void actionPerformed( ActionEvent e )
    {
        if ( e.getSource() == restartItem ) {
            buttonPressed( 2 );
        }
        else if ( e.getSource() == settingsItem ) {
            buttonPressed( 3 );
        }
        else if ( e.getSource() == saveItem ) {
            saveConsole();
        }
        else if ( e.getSource() == disconnectItem ) {
            disconnect();
        }
        else if ( e.getSource() == emailItem ) {
            buttonPressed( 4 );
        }
        else if ( e.getSource() == sinkItem ) {
            activateSink();
        }
        else if ( e.getSource() == toolbar5 ) {
            buttonPressed( 5 );
        }
        else if ( e.getSource() == toolbar6 ) {
            buttonPressed( 6 );
        }
        else {
            super.actionPerformed( e );
        }
    }


    /**
     *  Toolbar buttons
     */
    public void buttonPressed( int index )
    {
        String emailTitle;

        super.buttonPressed( index );

        switch ( index ) {

        case 1:
            /*
             *  Exit
             */
            shutDown();
            break;

        case 2:
            /*
             *  Restart
             */
            shutDownEngine();
            startUp();

            /*
             *  Update title for button 4
             */
            emailTitle = getEmailTitle();
            if ( emailTitle != null ) {
                setButtonLabel( 4, emailTitle );
            }
            break;

        case 3:
            /*
             *  Settings
             */
            settings();
            break;

        case 4:
            /*
             *  Start Email-App or bring to foreground
             */
            startEmailApp();
            break;

        case 5:
            /*
             *  Test StartApp (OPL)
             */
            debug( "Test 5" );
            try {
                int tid = EpocUtil.startApp( "OPL",
                                             null,
                                             "RC:\\System\\Opl\\TaskList.opo",
                                             EpocUtil.CMD_BACKGROUND );
                log( "ThreadId: " + tid );
            }
            catch ( Exception e ) {
                log( e );
                e.printStackTrace();
            }
            break;

        case 6:
            /*
             *  Test foregroundTask
             */
            debug( "Test 6" );
            int wgId = EpocUtil.windowGroupByName( 0, "*Jotter\0*" );
            int err = EpocUtil.foregroundWindowGroup( wgId );
            log( "foreground( \"*Jotter\\0*\" ): err = " + err );
            if ( err != 0 ) {
                log( "reason = " + EpocUtil.reason );
            }
            break;
        }
    }


    /**
     *  Create and start engine thread
     */
    public void startUp()
    {
        debug( "startUp" );
        engine = new SmtpAuth( out, args );
        thread = new Thread( engine );
        debug( "thread.start()" );
        thread.start();
    }


    /**
     *  Shut down engine and wait for thread to complete
     */
    public void shutDownEngine()
    {
        debug( "shutDown" );
        try {
            engine.shutDown();
            thread.join();
        }
        catch ( Exception e ) {}
    }


    /**
     *  Exit program - shut down engine before
     */
    public void shutDown()
    {
        /*
         *  shutdown engine
         */
        shutDownEngine();
        super.shutDown();
    }


    /**
     *  Save the console data
     */
    public void saveConsole()
    {
        try {
            File file = new File( System.getProperty( "user.home" ),
                                  SmtpAuth.NAME + "Console.txt" );
            PrintWriter w = new PrintWriter( new FileWriter( file ) );
            String contents[] = console.getContents();
            for ( int i = 0; i < contents.length; ++i ) {
                w.println( contents[ i ] );
            }
            w.close();
            log( "Console dumped to " + file.getCanonicalPath() );
        }
        catch ( Exception e ) {
            log( e );
        }
    }


    /**
     *  Disconnect from Internet
     */
    public void disconnect()
    {
        try {
            EpocUtil.netStop();
            log( "disconnected" );
        }
        catch ( Exception e ) {
            log( e );
        }
    }


    /**
     *  change settings
     */
    public void settings()
    {
        Thread thread = new Thread() {

            public void run()
            {
                try {
                    String editor =
                        engine.getIni().get( "Global", "editor", "Editor" );
                    log( "starting " + editor + " " + engine.iniName );
                    int tid = EpocUtil.startApp( editor, engine.iniName );
                    if ( EpocUtil.isDebug() ) {
                        log( "waiting for thread " + tid );
                    }
                    long now = System.currentTimeMillis();
                    int err = EpocUtil.waitThread( tid );
                    long delay = System.currentTimeMillis() - now;
                    if ( 0 == err ) {
                        if ( EpocUtil.isDebug() ) {
                            log( "switching " + SmtpAuth.NAME
                                        + " to foreground after "
                                        + delay + " ms" );
                        }
                        err = EpocUtil.foreground();
                    }
                    if ( 0 != err ) {
                        log( "error " + err );
                    }
                    if ( false ) {
                        /*
                         *  Update title for button 4
                         */
                        String emailTitle = getEmailTitle();
                        if ( emailTitle != null ) {
                            setButtonLabel( 4, emailTitle );
                        }
                    }
                }
                catch ( IOException ioe ) {
                    log( ioe );
                }
                /*
                 *  clean up thread data in DLL!
                 */
                EpocUtil.cleanup();
            }
        };
        thread.start();
    }


    /**
     *  find out about title of MsgApp (email)
     */
    public String getEmailTitle()
    {
        if ( !isEpoc ) {
            return null;
        }
        String title = System.getProperty( "epoc.msgapp.title" );
        String[] titles = new String[] { "Email", "E-Mail", title };
        for ( int i = titles.length - 1; i >= 0; --i ) {
            int wgId = EpocUtil.windowGroupByName( 0, titles[ i ] + "\0*" );
            if ( wgId >= 0 ) {
                /*
                 *  running app found
                 */
                title = titles[ i ];
                break;
            }
        }

        if ( title == null ) {
            /*
             *  app not running and not set in ini file, assume default
             */
            title = titles[ 0 ];
        }
        return title;
    }


    /**
     *  Activate Sink
     */
    public void activateSink()
    {
        Sink.activate();
        sinkItem.setEnabled( false );
    }



    /**
     *  Start MsgApp (email) or bring to foreground
     */
    public void startEmailApp()
    {
        String emailTitle = getEmailTitle();
        int wgId = EpocUtil.windowGroupByName( 0, emailTitle + "\0*" );
        if ( wgId >= 0 ) {
            int err = EpocUtil.foregroundWindowGroup( wgId );
            if ( EpocUtil.isDebug() ) {
                log( "foreground( \"" + emailTitle + "\\0*\" )"
                           + ": err = " + err );
                if ( err != 0 ) {
                    log( "reason = " + EpocUtil.reason );
                }
            }
        }
        else {
            /*
             *  start MsgApp
             */
            try {
                int tid = EpocUtil.startApp( "MsgApp", null, null,
                                             EpocUtil.CMD_OPEN );
                if ( EpocUtil.isDebug() ) {
                    log( "MsgApp started, ThreadId: " + tid );
                }
            }
            catch ( Exception e ) {
                log( "MsgApp not started" );
                log( e );
            }
        }
    }


    /**
     *  print a message
     */
    private void log( Object o )
    {
        String line = o.toString();
        if ( line.endsWith( "\n" ) ) {
            out.print( line );
        }
        else {
            out.println( line );
        }
        out.flush();
        debug( line );
    }


    /**
     *  Start application
     */
    public static void main( String[] args )
            throws Exception
    {
        if ( EpocUtil.isEpoc() ) {
            /*
             *  are we already running?
             */
            int wgId1 = EpocUtil.windowGroup();
            int wgId2 = EpocUtil.windowGroup( wgId1 );

            if ( wgId2 != -1 ) {
                /*
                 *  move older window to foreground
                 */
                debug( "1: " + wgId1 + ", 2:" + wgId2 );
                EpocUtil.foregroundWindowGroup( wgId2 );
                EpocUtil.cleanup();
                System.exit( 0 );
            }
        }

        /*
         *  new instance
         */
        SmtpAuthGui.defaultSizeTag = NORMAL;
        SmtpAuthGui gui = new SmtpAuthGui( args );
    }

}


