/****************************************************************************

Package:            mcutil
Class:              IniFile

Aufgabe:            Lesen und Schreiben von INI-Dateien (Windows like)

Programmierer:      Marcus von Cube, EasyData (MvC)

Aenderungen:

Datum     Name  Vers.  Bemerkung
-----------------------------------------------------------------------------
27.02.97  MvC   0.1    erstellt
01.04.97  MvC   1.0    In Package aufgenommen
09.12.99  MvC   1.1    JDK 1.1
20.04.00  MvC   1.2    getSection
16.05.03  MvC   2.0    Reorganisiert, neues Package
04.06.03  MvC   2.1    getBoolean
09.08.03  MvC   2.1a   getBoolean versteht auch "on"

****************************************************************************/

package de.mvcsys.util;

import java.io.File;
import java.io.PrintWriter;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

import java.util.*;

/**
 *  Lesen und Schreiben von INI-Dateien (Windows like)
 */
public class IniFile extends Liste
{
    /**
     *  Liste von Werten einer Section
     */
    public class Section extends Liste
    {
        private String name;

        /**
         *  erzeugt eine Section
         *  @param der Name der Section
         */
        Section( String name )
        {
            this.name = name;
        }


        /**
         *  Ausgabe der Section als String
         *  @return die Section als String mit Zeilentrennern NL
         */
        public String toString()
        {
            StringBuffer s = new StringBuffer( "[" + name + "]" );
            Enumeration e = keys();
            while ( e.hasMoreElements() ) {
                String key = (String) e.nextElement();
                Object[] triple = find( key );
                String value = (String) triple[ 1 ];
                if ( key.startsWith( ";" ) ) {
                    s.append( NL + value );
                }
                else if ( triple[ 2 ] == ENCODED ) {
                    s.append( NL + key + "~=" + value );
                }
                else {
                    s.append( NL + key + "=" + value );
                }
            }
            s.append( NL );
            return s.toString();
        }
    }


    /**
     *  Konstante Newline
     */
    public final static String NL = System.getProperty( "line.separator" );

    /*
     *  Variablen
     */
    private String filename;

    /**
     *  Konstruktor
     *  liest die komplette Ini-Datei ein
     */
    public IniFile( String filename ) throws IOException
    {
        this.filename = filename;
        File f = new File( this.filename );
        int line_no = 0;
        if ( f.exists() ) {
            BufferedReader d = new BufferedReader(
                                   new InputStreamReader(
                                       new FileInputStream( f ) ) );
            String line;
            String sname = "Global";
            do {
                line = d.readLine();
                if ( line != null ) {
                    ++line_no;
                    line = line.trim();
                    if ( line.startsWith( ";" ) ) {
                        put( sname, ";" + line_no, line );
                        continue;
                    }
                    if ( line.length() == 0 ) {
                        continue;
                    }
                    if ( line.startsWith( "[" ) ) {
                        /*
                         *  [Section]
                         */
                        int end = line.indexOf( "]" );
                        if ( end > 0 ) {
                            sname = line.substring( 1, end );
                        }
                        else {
                            sname = line.substring( 1 );
                        }
                        continue;
                    }
                    int gleich = line.indexOf( "=" );
                    int encoded = line.indexOf( "~=" );
                    if ( gleich <= 0 || gleich + 1 >= line.length() ) {
                        continue;
                    }
                    if ( encoded > 0 ) {
                        String var = line.substring( 0, encoded );
                        String val = line.substring( encoded + 2 );
                        put( sname, var, decode( val ), true );
                    }
                    else {
                        String var = line.substring( 0, gleich );
                        String val = line.substring( gleich + 1 );
                        put( sname, var, val );
                    }
                }
            } while ( line !=null );
            d.close();
        }
    }


    /**
     *  Sections und Werte einfuegen
     */
    public String put( String sname, String var, String wert )
    {
        return put( sname, var, wert, false );
    }


    /**
     *  Sections und Werte einfuegen
     */
    public String put( String sname, String var, String wert, boolean encoded )
    {
        Section s = (Section) get( sname );
        if ( s == null ) {
            /*
             *  neue Section einfuegen
             */
            s = new Section( sname );
            put( sname, s );
        }
        return (String) s.put( var, wert, encoded );
    }


    /**
     *  Section lesen
     *  @param sname Name der Section
     *  @return die Section als Liste
     */
    public Section getSection( String sname )
    {
        return (Section) get( sname );
    }


    /**
     *  Wert lesen
     *  @param sname Name der Section
     *  @param var Name der Variablen
     *  @param def Default
     *  @return Wert der Variable
     */
    public String get( String sname, String var, String def )
    {
        String result = null;
        Section s = getSection( sname );
        if ( s != null ) {
            result = (String) s.get( var );
        }
        return result == null ? def : result;
    }


    /**
     *  Wert lesen
     *  @param sname Name der Section
     *  @param var Name der Variablen
     *  @return Wert der Variable
     */
    public String get( String sname, String var )
    {
        return get( sname, var, null );
    }


    /**
     *  Wert als Zahl lesen
     *  @param sname Name der Section
     *  @param var Name der Variablen
     *  @param def Default
     *  @return Wert der Variable als Zahl
     */
    public int getInt( String sname, String var, int def )
    {
        String sresult = get( sname, var, null );
        int result = def;
        if ( sresult != null ) {
            try {
                result = Integer.parseInt( sresult );
            }
            catch ( Exception e ) {}
        }
        return result;
    }


    /**
     *  Wert als Zahl lesen
     *  @param sname Name der Section
     *  @param var Name der Variablen
     *  @return den Wert der Variable als Zahl
     */
    public int getInt( String sname, String var )
    {
        return getInt( sname, var, 0 );
    }


    /**
     *  Wert als Boolean lesen
     *  @param sname Name der Section
     *  @param var Name der Variablen
     *  @param def Default
     *  @return den Wert der Variable als Zahl
     */
    public boolean getBoolean( String sname, String var, boolean def )
    {
        String sresult = get( sname, var, null );
        boolean result = def;
        if ( sresult != null ) {
            try {
                result = 0 != Integer.parseInt( sresult );
            }
            catch ( Exception e ) {
                sresult = sresult.toLowerCase();
                result = sresult.equals( "true" )
                      || sresult.equals( "t" )
                      || sresult.equals( "yes" )
                      || sresult.equals( "y" )
                      || sresult.equals( "on" )
                      || sresult.equals( "ja" )
                      || sresult.equals( "j" );
            }
        }
        return result;
    }


    /**
     *  Wert als Boolean lesen
     *  @param sname Name der Section
     *  @param var Name der Variablen
     *  @return den Wert der Variable als Zahl
     */
    public boolean getBoolean( String sname, String var )
    {
        return getBoolean( sname, var, false );
    }


    /**
     *  Ausgabe der gesamten Datei als Text
     *  @return die Datei mit allen Kommentaren und Sections mit NL als
     *          Zeilentrenner
     */
    public String toString()
    {
        StringBuffer str = new StringBuffer();
        Enumeration e = keys();
        boolean first = true;
        while ( e.hasMoreElements() ) {
            if ( first ) {
                first = false;
            }
            else {
                str.append( NL );
            }
            str.append( get( e.nextElement() ) );
        }
        return str.toString();
    }


    /**
     *  Abspeichern der Datei unter ihrem aktuellen Namen
     */
    public void save() throws IOException
    {
        save( filename );
    }


    /**
     *  Abspeichern der Datei unter einem neuen Namen
     *  @param fname der Dateiname
     */
    public void save( String fname ) throws IOException
    {
        filename = fname;
        File f = new File( filename );
        if ( f.exists() ) {
            String bak;
            int pos = filename.lastIndexOf( "." );
            if ( pos > 0 ) {
                bak = filename.substring( 0, pos ) + ".bak";
            }
            else {
                bak = "~" + filename;
            }
            new File( bak ).delete();
            f.renameTo( new File( bak ) );
        }
        PrintWriter w = new PrintWriter( new FileOutputStream( fname ) );
        w.print( this + NL );
        w.close();
    }


    /**
     *  Test-main
     *  @param args Argumente: Name der INI-Datei
     */
    public static void main( String args[] )
    {
        IniFile inif = null;
        try {
            String name;
            if ( args.length > 0 ) {
                name = args[ 0 ];
            }
            else {
                name = "IniFile.ini";
            }
            inif = new IniFile( name );
            System.out.println(
                    inif.put( "Global", "geheim", "geheim", true ) );
            System.out.println( inif );
            System.out.println( "geheim=" + inif.get( "Global", "geheim" ) );
            inif.save();
        }
        catch ( IOException e ) {
            System.out.println( e );
        }
    }

}

 

